/*	Application		IPC receiver-side using socket (UDP) - Message serialisation example		
	File			IPC_Receiver_MessageSerialisation.java
	Purpose			Demonstration of local IPC, with a serialised data structure, using a blocking socket (on the receiving side)
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates a socket, and binds to local address (implicitly uses INADDR_ANY)
                                2. Waits for a message to arrive on its port (8007)
                                3. Displays the message 
                                4. Exits once a key is pressed

    To run the project from the command line, go to the dist folder and type: java -jar "IPC_Receiver_MessageSerialisation.jar"
*/
package ipc_receiver_messageserialisation;

import java.io.Console;
import java.net.*;
import java.nio.*; // java.nio.ByteBuffer

public class IPC_Receiver_MessageSerialisation
{
    private static int RECEIVE_PORT = 8007;
    private static DatagramSocket m_IPC_ReceiveSOCKET;
    private static InetAddress m_SenderIPAddress;   // Will hold the address and port of the sender after a message has been received

    private static int MAX_NAME_LENGTH = 20;
    private static int MAX_ADDRESS_LENGTH = 50;
    
    // Message structure (this will be deserialised from the received message before displaying contents)
    // To implement the equivallent of a C/C++/C# structure, in Java a class is used
    final class MyDetailsMessage {
        byte[] cName;       // (c++)  char cName[MAX_NAME_LENGTH];
        int iAge;           // 32-bit integer (C# Int32)
        byte[] cAddress;    // (c++)  char cAddress[MAX_ADDRESS_LENGTH];
    }; 

    private static int MyDetailsMessage_size = 74; // The size when serialised as a byte array
     
    static Console con;
    
    public static void main(String[] args)
    {
        con = System.console();
        if (null == con)
        {
            System.out.println("!! Error accessing console: Run in a command window !!");
            System.exit(-1);
        }

        System.out.println("          Local Inter-Process Communication (receiver side).");
        System.out.println("  This program waits for a message to arrive on port 8007 (decimal).");
        System.out.println("The message is displayed and the process exits when a key is pressed.");
        System.out.println("----------------------------------------------------------------------");
        System.out.println("Waiting for message ...   ");

        IPC_Receiver_MessageSerialisation Socket_Receiver = new IPC_Receiver_MessageSerialisation();
    }

    IPC_Receiver_MessageSerialisation()
    {
        CreateReceiveSocket();

        // Wait for, and receive the message
        byte[] bArray = ReadMessageFromReceiveBufferOrWait();
        MyDetailsMessage RecdMessage = DeSerialize_IPC_Reply(bArray);

        String sName = new String(RecdMessage.cName);
        String sAddress = new String(RecdMessage.cAddress);
        System.out.printf("   Name: %s\n   Address: %s\n   Age: %d\n",
                                    sName, sAddress, RecdMessage.iAge);
        CloseSocketAndExit();
    }

    void CreateReceiveSocket()
    {
        try
        {
            m_IPC_ReceiveSOCKET = new DatagramSocket(RECEIVE_PORT); // Create socket and bind it to the receive port (uses the local address)
            m_IPC_ReceiveSOCKET.setSoTimeout(0); // A timeout value of 0 is interpreted as infinity i.e blocking-IO behaviour
        }
        catch (SocketException ex)  // Catch exceptions from the socket operations
        {
            System.out.printf("Error %s\n", ex.toString());
            CloseSocketAndExit();
        }
    }

    byte[] ReadMessageFromReceiveBufferOrWait()
    {
        byte[] bRecvBuf = new byte[MyDetailsMessage_size];
        try
        {
            DatagramPacket RecvPacket = new DatagramPacket(bRecvBuf, MyDetailsMessage_size);
            m_IPC_ReceiveSOCKET.receive(RecvPacket);

            // Place the sender's address into the m_SenderIPAddress
            m_SenderIPAddress = RecvPacket.getAddress();
        }
        catch(Exception ex)
        {
            System.out.printf("Receive failed: %s\n", ex.toString());
            CloseSocketAndExit();
        }
        return bRecvBuf;
    }

    public MyDetailsMessage DeSerialize_IPC_Reply(byte[] bRecdByteArray)
    {   // Converts a byte array (from Receive) to a MyDetailsMessage structure (Fields have fixed length, but non-fixed data length)
        MyDetailsMessage result_MyDetailsMessage = new MyDetailsMessage();
        result_MyDetailsMessage.cName = new byte[MAX_NAME_LENGTH];
        result_MyDetailsMessage.cAddress = new byte[MAX_ADDRESS_LENGTH];
        try
        {
            ByteBuffer bbuff = ByteBuffer.wrap(bRecdByteArray);
            bbuff.position(0); // Start of first field
            bbuff.get(result_MyDetailsMessage.cName, 0/*offset into dest array*/, MAX_NAME_LENGTH/*length*/);
            result_MyDetailsMessage.iAge = ByteArrayToInt(bRecdByteArray, MAX_NAME_LENGTH/*Start of second field*/);
            bbuff.position(0 + MAX_NAME_LENGTH + 4); // Start of third field
            bbuff.get(result_MyDetailsMessage.cAddress, 0/*offset into dest array*/, MAX_ADDRESS_LENGTH/*length*/);
        } 
        catch (Exception ex)
        {
            System.out.printf("DeSerialize failed: %s\n", ex.toString());
        }
        return result_MyDetailsMessage;
    }
     
    int ByteArrayToInt(byte[] bArray, int iStartIndex)
    {
        int iValue = bArray[iStartIndex] + (bArray[iStartIndex + 1] * 256);
            iValue += (bArray[iStartIndex + 2] * 256 * 256)+ (bArray[iStartIndex + 3]* 256 * 256 * 256);
        return iValue;
    }

    void CloseSocketAndExit()
    {
        try
        {
            m_IPC_ReceiveSOCKET.close();
            System.out.println("Socket closed");
        }
        catch(Exception ex)
        {
        }
        System.out.println("(press a key to continue)");
        con.readLine();	// Wait for user to type a key before the program exits
        System.exit(1);
    }
}