/*	Application		IPC sender-side using socket (UDP) - Message serialisation example		
	File			IPC_Sender_MessageSerialisation.java
	Purpose			Demonstration of local IPC, with a serialised data structure, using a blocking socket (on the receiving side)
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates a socket
                                2. Waits for user to type a message
                                3. Sends the message to the the application port (8007) on the local computer
                                4. Exits once the message has been sent

    To run the project from the command line, go to the dist folder and type: java -jar "IPC_Sender_MessageSerialisation.jar"
*/
package ipc_sender_messageserialisation;

import java.io.Console;
import java.net.*;
import java.nio.*; // java.nio.ByteBuffer

public class IPC_Sender_MessageSerialisation
{
    private static int SEND_PORT = 8007;
    // Use a different port for binding to, so both sender and receiver processes can run on the same computer
    private static  int LOCAL_PORT = 8000; 
    private static DatagramSocket m_IPC_SendSOCKET;
    private static InetAddress m_DestinationIPAddress;   // Will hold the address and port of the sender after a message has been received

    private static int MAX_NAME_LENGTH = 20;
    private static int MAX_ADDRESS_LENGTH = 50;
    
    // Message structure (this will be deserialised from the received message before displaying contents)
    // To implement the equivallent of a C/C++/C# structure, in Java a class is used
    final class MyDetailsMessage {
        byte[] cName;       // (c++)  char cName[MAX_NAME_LENGTH];
        int iAge;           // 32-bit integer (C# Int32)
        byte[] cAddress;    // (c++)  char cAddress[MAX_ADDRESS_LENGTH];
    }; 

    private static int MyDetailsMessage_size = 74; // The size when serialised as a byte array
     
    static Console con;
    
    public static void main(String[] args)
    {
        con = System.console();
        if (null == con)
        {
            System.out.println("!! Error accessing console: Run in a command window !!");
            System.exit(-1);
        }

        System.out.println("Local Inter-Process Communication (sender side).");
        System.out.println("This program waits for the user to type a message,");
        System.out.println("and sends it to port 8007 on the local computer.");
        System.out.println("--------------------------------------------------");
        System.out.println("Waiting for message ...   ");

        IPC_Sender_MessageSerialisation Socket_Sender = new IPC_Sender_MessageSerialisation();
    }

    IPC_Sender_MessageSerialisation()
    {
        CreateSendSocket();

         MyDetailsMessage Message_PDU = new MyDetailsMessage();
        // The message to be sent will be held in the form of a MyDetailsMessage structure.
        // The data values are written into the structure using the specific field names
        // (which are cName, cAddress and iAge).
        System.out.println("Customer name:		");
        try
        {
            Message_PDU.cName = con.readLine().getBytes("US-ASCII");
        }
        catch(Exception Ex)
        {
        }
        System.out.println("Customer address:	");
        try
        {
            Message_PDU.cAddress = con.readLine().getBytes("US-ASCII");
        }
        catch(Exception Ex)
        {
        }
        System.out.println("Customer age:		");
        try
        {
            Message_PDU.iAge = Integer.parseInt(con.readLine());
        }
        catch (Exception ex)
        {
            Message_PDU.iAge = 0;
            System.out.println("!! Invalid value (Customer age) !!");
        }

        // Send the message that the user typed, to the specified port on the local computer
        SendMessage(Message_PDU);
        CloseSocketAndExit();
    }

    void CreateSendSocket()
    {
        try
        {
            m_IPC_SendSOCKET = new DatagramSocket(LOCAL_PORT); // Create socket and bind it to the provided port
            m_IPC_SendSOCKET.setSoTimeout(0); // A timeout value of 0 is interpreted as infinity i.e blocking-IO behaviour
        }
        catch (SocketException ex)  // Catch exceptions from the socket operations
        {
            System.out.printf("Error %s\n", ex.toString());
            CloseSocketAndExit();
        }
    }
    
    void SendMessage(MyDetailsMessage Message)
    {
        byte[] bSendBuf = Serialize_Message_PDU(Message);
        try
        {
            InetAddress SendIPAddress = InetAddress.getLoopbackAddress(); // Assume receiver is at same address as sender
            DatagramPacket SendPacket = new DatagramPacket(bSendBuf, MyDetailsMessage_size,
                                    SendIPAddress /*The address to send to*/, SEND_PORT); 
            m_IPC_SendSOCKET.send(SendPacket);
        }
        catch(Exception ex)
        {
            System.out.printf("Send failed: %s\n", ex.toString());
            CloseSocketAndExit();
        }
    }

    private byte[] Serialize_Message_PDU(MyDetailsMessage Message)
    {   // Convert an MyDetailsMessage structure into a byte array
        byte[] bBuf = new byte[MyDetailsMessage_size];
        ByteBuffer bbWrapped = ByteBuffer.wrap(bBuf);
        byte[] bAge = IntToByteArray(Message.iAge);
        bbWrapped.position(0/*Start of field 1*/);
        bbWrapped.put(Message.cName);      
        bbWrapped.position(MAX_NAME_LENGTH/*Start of field 2*/);
        bbWrapped.put(bAge);
        bbWrapped.position(MAX_NAME_LENGTH + 4/*Start of field 3*/);
        bbWrapped.put(Message.cAddress);
        return bBuf;
    }
     
    int ByteArrayToInt(byte[] bArray, int iStartIndex)
    {
        int iValue = bArray[iStartIndex] + (bArray[iStartIndex + 1] * 256);
            iValue += (bArray[iStartIndex + 2] * 256 * 256)+ (bArray[iStartIndex + 3]* 256 * 256 * 256);
        return iValue;
    }
    
    byte[] IntToByteArray(int iValue)
    {
        byte[] bArray = new byte[4];
        bArray[3] = (byte)(iValue / (256 * 256 * 256));
        bArray[2] = (byte)((iValue - (bArray[3] * 256 * 256 * 256)) / (256 * 256));
        bArray[1] = (byte)((iValue - ((bArray[3] * 256 * 256 * 256) + (bArray[2] * 256 * 256))) / 256);
        bArray[0] = (byte)(iValue % 256); // Remainder
        return bArray;
    }

    void CloseSocketAndExit()
    {
        try
        {
            m_IPC_SendSOCKET.close();
            System.out.println("Socket closed");
        }
        catch(Exception ex)
        {
        }
        System.out.println("(press a key to continue)");
        con.readLine();	// Wait for user to type a key before the program exits
        System.exit(1);
    }
}